/******************************************************************************
                               Include Files
 ******************************************************************************/

#include <mpi.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
 
/******************************************************************************
                     Definition of Constants & Functions
 ******************************************************************************/

#define Max_integer 32767
#define Max_long    2147483647

#define success   1
#define fail      0

#define SOURCE    0
#define TARGET    1
#define POINTER   2

#define SPARSE    1
#define DENSE     0

#define null         0
#define vacant      -1		/* Hashtable or buffer is vacant */
#define self_index  -1		/* Pointing to myself */
#define empty     0

#define Rankmax   1		/* Maximum Rank of input array */
#define Nmax      2097152 	/* Maximum global domain size  */
#define Lmax      8192		/* Maximum local domain size   */
#define Procmax   128		/* Maximum number of processor on each dim*/
#define Tot_Procmax 128		/* Maximum total number of processor */

#define rport     8

#define Htablemax (int) (Lmax*1.5)
#define Indexmax  Nmax

#define TAG       3
#define intsize   4
#define datasize  4

/*
 #define abs(x)      ( (x) < 0     ?  -(x)  :  (x) )
*/
#define Max(x,y)    ( (x) < (y)   ?   (y)  :  (x) )
#define Min(x,y)    ( (x) < (y)   ?   (x)  :  (y) )

#define Ceiling(j,k)   ( ( (j) + (k) - 1 ) / (k) )
#define Floor(j,k)     ( (j) / (k) )

/* #define msg_size      4 */       /* integer */
/* #define datasize      4 */       /* integer */

#define Hot_Proc_Max  Tot_Procmax/2

#define master   0

/******************************************************************************
                     Definition of Type
 ******************************************************************************/
 
typedef int Gindex;   /* for global index */

typedef int Lindex;   /* for local index */

typedef int Pindex;   /* for processor index */

typedef enum {false, true} 
    Boolean;

/*----------------------*/

typedef int Data_Type; 

/*----------------------*/

typedef struct {
  Data_Type data;
  int ptr;
} Buffer_Type;

typedef struct {
  int ptr;
  int value;
} Hash_table_type;


/******************************************************************************
                     Definition of Global variables
 ******************************************************************************/
 
MPI_Comm  ORI_WORLD, WORLD;
MPI_Group ORI_WORLD_GROUP, WORLD_GROUP;
int exclude[Tot_Procmax];

int s_rank;               /* Rank of source array */
int t_rank;               /* Rank of target array */

int ori_nproc;          /* Number of available  processors */

int tot_nproc;          /* max(s_tot_nproc, t_tot_nproc) */

int  namelen;
char processor_name[MPI_MAX_PROCESSOR_NAME];

int only_master;

/*-----------------------------------------------------------------*/
/*-----------------------  Source Distribution  -------------------*/
/*-----------------------------------------------------------------*/

int s_N[Rankmax],       /* Global Domain size */
    s_tot_nproc,        /* Total number of source processors */
    s_nproc[Rankmax],   /* Number of source processors */
    s_width[Rankmax],   /* Width of partition within a single tile */
                        /* that is, block size */
    s_twidth[Rankmax],  /* Width of each tile */
    s_ntile[Rankmax],   /* Total Number of tiles */
    s_tnum[Rankmax],    /* Tile number */
    s_L[Rankmax];       /* Local domain size for myself */

int s_N_all, s_L_all;   /* Total global & local size  for myself */

int s_L_table[Rankmax][Tot_Procmax]; /* Local domain sizees on all processors */
int s_L_all_table[Tot_Procmax]; /* Total local domain sizees on all processors */
int s_L_max[Rankmax];                  /* Maximum local size (probably on P0) */
int s_L_all_max;                       /* Maximum local size (probably on P0) */

int my_s_grid[Rankmax];            /* My source pnum of grid form */
int s_grid[Tot_Procmax][Rankmax];  /* Each source pnums of grid form */
 
Boolean  s_active;                 

Buffer_Type  *sendB;

int sendl[Tot_Procmax];            /* Row vector of comm matrix */
int sendl_self;
int send_offset[Tot_Procmax];

int tot_sendl;


/*-----------------------------------------------------------------*/
/*-------------------  Target Distribution  -----------------------*/
/*-----------------------------------------------------------------*/

int t_N[Rankmax],       /* Global Domain size */
    t_tot_nproc,        /* Total number of source processors */
    t_nproc[Rankmax],   /* Number of target processors */
    t_width[Rankmax],   /* Width of partition within a single tile */
                        /* that is, block size */
    t_twidth[Rankmax],  /* Width of each tile */
    t_ntile[Rankmax],   /* Total Number of tiles */
    t_tnum[Rankmax],    /* Tile number */
    t_L[Rankmax];       /* Local domain size for myself */

int t_N_all, t_L_all;   /* Total global & local size for myself */

int t_L_table[Rankmax][Tot_Procmax]; /* Local domain sizees on all processors */
int t_L_all_table[Tot_Procmax]; /* Total local domain sizees on all processors */
int t_L_max[Rankmax];                  /* Maximum local size (probably on P0) */
int t_L_all_max;                       /* Maximum local size (probably on P0) */

int my_t_grid[Rankmax];            /* My target pnum of grid form */
int t_grid[Tot_Procmax][Rankmax];  /* Each target pnums of grid form */

Boolean  t_active;
Boolean  BLOCK;

Buffer_Type *recvB;

int recvl[Tot_Procmax];            /* Column vector of comm matrix */
int recv_offset[Tot_Procmax];
int tot_recvl;


/*-----------------------------------------------------------------*/
/*-----------------------  General Distribution -------------------*/
/*-----------------------------------------------------------------*/

int my_pnum;                /* Processor number */
int ori_my_pnum;            /* Processor number */

Boolean enable;         /* Enable or Disabel */
   
Data_Type S0[Lmax],           /* Source array */
          T0[Lmax];           /* Target Array */

int       P0[Lmax];		/* Pointer Array */

int       S0_lidx[Lmax];	
int       P0_lidx[Lmax];		
int       P0_pnum[Lmax];		

/****************** rank -2
int S1[Lmax][Lmax],     
    T1[Lmax][Lmax],     
    P1[Lmax][Lmax];	
****************************/

/*-----------------------------------------------------------------*/
/*-------------------------------  Bucket  ------------------------*/
/*-----------------------------------------------------------------*/

int K_value;

double local_dense_factor;
double global_dense_factor;


/******************************************************************************
                            Definition of Timer
 ******************************************************************************/

#define  COMMON_setup      	0

#define  OSC_tot       		1
#define  OSC_lcomp_tot          2
#define  OSC_comm      		3
#define  OSC_scan1     		4
#define  OSC_scan2     		5
#define  OSC_decomp    		6

#define  TSC_raw_tot                    10
#define  TSC_stage1_tot                 11
#define  TSC_stage2_tot                 12
#define  TSC_raw_set                    13
#define  TSC_raw_scan1                  14
#define  TSC_raw_lcomp_stage1           15
#define  TSC_raw_lcomp_stage2           16
#define  TSC_raw_comm_stage1            17
#define  TSC_raw_comm_stage2            18
#define  TSC_raw_comm_global_stage1     19
#define  TSC_raw_comm_global_stage2     20
 
#define  MSC_tot       		31
#define  MSC_comm      		32
#define  MSC_lcomb     		33
#define  MSC_lgp     		34
#define  MSC_lcopy    		35

/*-----------------------------------------------------------------*/
/*---------------------------  Hash Table  ------------------------*/
/*-----------------------------------------------------------------*/

/*Hash_table_type Htable[(int) (Lmax*1.5)];
*/

Hash_table_type Htable[Lmax];
int Htable_size;

